/*  $Id: SortedStreamMerger.java 1335 2012-08-25 13:27:49Z sdevaux $
 *
 * Tabs = 4 chars
 * copyright (C) 2002-2012 Sebastien Devaux
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 *              Sebastien Devaux <sebastien.devaux@laposte.net>
 */
package net.eduvax.util;

import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.util.Collections;
import java.util.Comparator;
import java.util.Vector;

/** Fusion de streams ordonnés tout en préservant l'ordre dans le 
 * résultat. 
 * <p>Peut être utilisé pour l'étape finale de fusion d'un tri externe
 * (cf littérature à propos des "external sort algorithms")</p>
 * <p><b>Attention</b>, ceci ne fonctionne évidemment que si les entrées à
 * fusionner sont déjà triées selon le même critère de tri que celui qui
 * sera appliqué pendant la fusion.</p>
 */ 
public class SortedStreamMerger implements Runnable
{
    /** Créer nouveau fusionneur */
    public SortedStreamMerger()
    {
        _inputs=new Vector<LineReaderBuffer>();
        _errHandler=App.get().getErrHandler(ErrHandlerFactory.M_NONE);
    }
    /** Ajouter un stream d'entrée à fusionner 
     * @param str Stream à ajouter
     */ 
    public void addStream(InputStream str) 
    {
            _inputs.add(new LineReaderBuffer(str));
    }    
    /** Ajouter un fichier d'entrée à fusionner. 
     * @param path Chemin du fichier à ajouter.
     */ 
    public void addFile(String path)
    {
        try {
            _inputs.add(new LineReaderBuffer(path));
        }
        catch (IOException ex) {
            Error err=new Error(getClass().getName()+".addin",Error.ELevel.ERR,ex);
            err.addArg(this);
            _errHandler.handle(err);
        }
    }
    /** Définir le flux de sortie. 
     * @param s flux destination de la fusion.
     */ 
    public void setOutputStream(OutputStream s)
    {
        _output=new PrintStream(s);
    }
    /** Définir le fichier de sortie.
     * @param path Chemin du fichier cible.
     */ 
    public void setOutputPath(String path)
    {
        try {
            _output=new PrintStream(new FileOutputStream(path));
        }
        catch (IOException ex) {
            Error err=new Error(getClass().getName()+".setout",Error.ELevel.ERR,ex);
            err.addArg(this);
            _errHandler.handle(err);
            _output=System.err;
        }
    }
    /** Mise en oeuvre Runnable. */
    public void run()
    {
        // Ce serait mieux avec une collection triée mais
        // le framework .Net 2 n'a pas ça, il y a Sortedset qui
        // pourrait faire l'affaire dans .Net 4.
        // La compatibilité .Net 2 reste nécessaire avec mono (pour
        // LogSupSESAM51)
        try {
            Vector<LineReaderBuffer> list=new Vector<LineReaderBuffer>();
            Comparator<LineReaderBuffer> comp=new DefaultLineBufferComparator();
            for(LineReaderBuffer buf : _inputs)
            {
                buf.readNext();
                list.add(buf);
            }
            Collections.sort(list,comp);
            while (list.size()>0) 
            {
                LineReaderBuffer buf=list.elementAt(0);
                list.remove(buf);
                _output.println(buf.peek());
                if (buf.readNext()!=null) 
                {
                    list.add(buf);
                    Collections.sort(list,comp);
                }
                else 
                {
                    buf.close();
                }
            }
            _output.close();
        }
        catch (IOException ex) {
            Error err=new Error(getClass().getName()+".io",Error.ELevel.ERR,ex);
            err.addArg(this);
            _errHandler.handle(err);
        }
    }
    /** */
    private ErrorHandler _errHandler;
    /** Chemin fichier de sortie. */
    private PrintStream _output;
    /** Chemins fichiers d'entrée. 
     * @djin.relation composition,-,*,LineReaderBuffer
     */ 
    private  Vector<LineReaderBuffer> _inputs;
    /** Comparateur de ligne par defaut. */
    class DefaultLineBufferComparator implements Comparator<LineReaderBuffer> {
        /** Comparer deux ligne.
         * @param x Première élément de la comparaison.
         * @param y Deuxième élément de la comparaison.
         * @return Code comparaison selon spec interface IComparer.
         */ 
        public int compare(LineReaderBuffer x,LineReaderBuffer y) {
            return x.peek().compareTo(y.peek());
        }
    }
}
